from typing import TYPE_CHECKING, Type

import isaaclab.utils.math as math_utils
import torch
from isaaclab.envs.mdp.actions.actions_cfg import (
    DifferentialInverseKinematicsActionCfg as __DifferentialInverseKinematicsActionCfg,
)
from isaaclab.envs.mdp.actions.task_space_actions import (
    DifferentialInverseKinematicsAction as __DifferentialInverseKinematicsAction,
)
from isaaclab.managers.action_manager import ActionTerm
from isaaclab.utils import configclass

if TYPE_CHECKING:
    from srb._typing import AnyEnv
    from srb.core.asset import Articulation


class DifferentialInverseKinematicsAction(__DifferentialInverseKinematicsAction):
    cfg: "DifferentialInverseKinematicsActionCfg"
    _env: "AnyEnv"
    _asset: "Articulation"

    def __init__(self, cfg: "DifferentialInverseKinematicsActionCfg", env: "AnyEnv"):
        super().__init__(
            cfg,
            env,  # type: ignore
        )

        if self.cfg.base_name:
            base_ids, base_names = self._asset.find_bodies(self.cfg.base_name)
            if len(base_ids) != 1:
                raise ValueError(
                    f"Expected one match for the base name: {self.cfg.base_name}. Found {len(base_ids)}: {base_names}."
                )
            self._base_idx = base_ids[0]
        else:
            self._base_idx = None

    @property
    def jacobian_b(self) -> torch.Tensor:
        jacobian = self.jacobian_w

        if self._base_idx:
            base_quat_w = self._asset.data.body_quat_w[:, self._base_idx]
        else:
            base_quat_w = self._asset.data.root_quat_w

        base_rot_matrix = math_utils.matrix_from_quat(math_utils.quat_inv(base_quat_w))
        jacobian[:, :3, :] = torch.bmm(base_rot_matrix, jacobian[:, :3, :])
        jacobian[:, 3:, :] = torch.bmm(base_rot_matrix, jacobian[:, 3:, :])
        return jacobian

    def _compute_frame_pose(self) -> tuple[torch.Tensor, torch.Tensor]:
        ee_pos_w = self._asset.data.body_pos_w[:, self._body_idx]
        ee_quat_w = self._asset.data.body_quat_w[:, self._body_idx]

        if self._base_idx is not None:
            base_pos_w = self._asset.data.body_pos_w[:, self._base_idx]
            base_quat_w = self._asset.data.body_quat_w[:, self._base_idx]
        else:
            base_pos_w = self._asset.data.root_pos_w
            base_quat_w = self._asset.data.root_quat_w

        ee_pose_b, ee_quat_b = math_utils.subtract_frame_transforms(
            base_pos_w, base_quat_w, ee_pos_w, ee_quat_w
        )

        if self.cfg.body_offset is not None:
            ee_pose_b, ee_quat_b = math_utils.combine_frame_transforms(
                ee_pose_b, ee_quat_b, self._offset_pos, self._offset_rot
            )

        return ee_pose_b, ee_quat_b

    def _compute_frame_jacobian(self) -> torch.Tensor:
        jacobian = self.jacobian_b
        if self.cfg.body_offset is not None:
            jacobian[:, :3, :] += torch.bmm(
                math_utils.skew_symmetric_matrix(self._offset_pos), jacobian[:, 3:, :]
            )
            jacobian[:, 3:, :] = torch.bmm(
                math_utils.matrix_from_quat(self._offset_rot), jacobian[:, 3:, :]
            )
        return jacobian


@configclass
class DifferentialInverseKinematicsActionCfg(__DifferentialInverseKinematicsActionCfg):
    class_type: Type[ActionTerm] = DifferentialInverseKinematicsAction
    base_name: str = ""



import subprocess
from importlib.util import find_spec
from typing import Iterable, Literal

from srb.utils import logging


def terminate_process(
    cmd: str | Iterable[str],
    process: subprocess.Popen | None = None,
    signal: Literal[
        "INT",
        "QUIT",
        "ABRT",
        "KILL",
        "TERM",
        "STOP",
    ] = "TERM",
):
    terminate_cmd = (
        "pkill",
        "--signal",
        signal,
        "--full",
        " ".join(cmd) if isinstance(cmd, Iterable) else cmd,
    )

    try:
        subprocess.run(terminate_cmd, check=False)

        if not process or find_spec("psutil") is None:
            return
        import psutil

        if not psutil.pid_exists(process.pid):
            return

        parent = psutil.Process(process.pid)
        for child in parent.children(recursive=True):
            if child.is_running():
                child.terminate()
        if parent.is_running():
            parent.terminate()

        for child in parent.children(recursive=True):
            if child.is_running():
                child.kill()
        if parent.is_running():
            parent.kill()
    except Exception as e:
        logging.error(f"Failed to terminate process: {e}")
    finally:
        subprocess.run(terminate_cmd, check=False)



from pathlib import Path

# Path to repository root directory
SRB_DIR = Path(__file__).resolve().parent.parent.parent

# Path to apps (experience) directory
SRB_APPS_DIR = SRB_DIR.joinpath("apps")

# Path to assets directory
SRB_ASSETS_DIR = SRB_DIR.joinpath("assets")
SRB_ASSETS_DIR_SRB = SRB_ASSETS_DIR.joinpath("srb_assets")
SRB_ASSETS_DIR_SRB_OBJECT = SRB_ASSETS_DIR_SRB.joinpath("object")
SRB_ASSETS_DIR_SRB_ROBOT = SRB_ASSETS_DIR_SRB.joinpath("robot")
SRB_ASSETS_DIR_SRB_SCENERY = SRB_ASSETS_DIR_SRB.joinpath("scenery")
SRB_ASSETS_DIR_SRB_SKYDOME = SRB_ASSETS_DIR_SRB.joinpath("skydome")
SRB_ASSETS_DIR_SRB_SKYDOME_LOW_RES = SRB_ASSETS_DIR_SRB_SKYDOME.joinpath("low_res")
SRB_ASSETS_DIR_SRB_SKYDOME_HIGH_RES = SRB_ASSETS_DIR_SRB_SKYDOME.joinpath("high_res")

# Path to hyperparameters directory
SRB_HYPERPARAMS_DIR = SRB_DIR.joinpath("hyperparams")

# Path to logs
SRB_LOGS_DIR = SRB_DIR.joinpath("logs")

# Path to a cached list of registered entities
SRB_CACHE_PATH: Path = SRB_DIR.joinpath(".cache")
SRB_ENV_CACHE_PATH: Path = SRB_CACHE_PATH.joinpath("env.json")
SRB_OBJECT_CACHE_PATH: Path = SRB_CACHE_PATH.joinpath("object.json")
SRB_ROBOT_CACHE_PATH: Path = SRB_CACHE_PATH.joinpath("robot.json")
SRB_SCENERY_CACHE_PATH: Path = SRB_CACHE_PATH.joinpath("scenery.json")



from dataclasses import MISSING
from typing import Callable, Tuple

from pxr import Usd

from srb.core.sim import SpawnerCfg
from srb.utils.cfg import configclass

from .impl import spawn_particles_grid, spawn_particles_pyramid


@configclass
class ParticlesSpawnerCfg(SpawnerCfg):
    # Distribution
    particle_size: float = 0.01
    particle_spacing: float = 0.0
    ratio: float = 1.0

    ## Particle system
    particle_system_enabled: bool = True
    particle_contact_offset: float | None = None
    max_velocity: float | None = None
    enable_ccd: bool = False
    wind: Tuple[float, float, float] = (0.0, 0.0, 0.0)
    # Advanced
    contact_offset: float | None = None
    rest_offset: float | None = None
    solid_rest_offset: float | None = None
    fluid_rest_offset: float | None = None
    solver_position_iterations: float = 16
    max_depenetration_velocity: float | None = None
    max_neighborhood: float = 96
    neighborhood_scale: float = 1.01
    global_self_collision_enabled: bool | None = None
    non_particle_collision_enabled: bool | None = None

    ## Particle set
    particle_enabled: bool = True
    self_collision: bool = True
    fluid: bool = True
    # Advanced
    particle_group: int = 0

    ## PBD Material
    density: float = 0.0
    friction: float = 0.2
    damping: float = 0.0
    viscosity: float = 0.0
    cohesion: float = 0.0
    surface_tension: float = 0.0
    drag: float = 0.0
    lift: float = 0.0
    # Advanced
    adhesion: float = 0.0
    particle_adhesion_scale: float = 1.0
    adhesion_offset_scale: float = 0.0
    particle_friction_scale: float = 1.0
    vorticity_confinement: float = 0.0
    cfl_coefficient: float = 1.0
    gravity_scale: float = 1.0

    ## Visuals
    cast_shadows: bool = False

    ## Particle anisotropy (fluid only)
    particle_anisotropy_enabled: bool = True
    anisotropy_scale: float = 1.0
    anisotropy_min: float = 0.2
    anisotropy_max: float = 2.0

    ## Particle smoothing (fluid only)
    particle_smoothing_enabled: bool = True
    smoothing_strength: float = 0.8

    ## Particle isosurface (fluid only)
    isosurface_enabled: bool = False
    isosurface_max_vertices: int = 2**20
    isosurface_max_triangles: int = 2**21
    isosurface_max_subgrids: int = 2**11
    isosurface_grid_spacing: float | None = None
    isosurface_surface_distance: float | None = None
    isosurface_mesh_smoothing_passes: int = 4
    isosurface_mesh_normal_smoothing_passes: int = 4
    isosurface_grid_filtering_passes: str = "GSRS"
    isosurface_grid_smoothing_radius: float | None = None


@configclass
class GridParticlesSpawnerCfg(ParticlesSpawnerCfg):
    func: Callable[..., Usd.Prim] = spawn_particles_grid

    # Distribution
    dim_x: int = MISSING  # type: ignore
    dim_y: int = MISSING  # type: ignore
    dim_z: int = MISSING  # type: ignore

    # Initial velocity
    velocity: Tuple[
        float | Tuple[float, float],
        float | Tuple[float, float],
        float | Tuple[float, float],
    ] = (0.0, 0.0, 0.0)


@configclass
class PyramidParticlesSpawnerCfg(ParticlesSpawnerCfg):
    func: Callable[..., Usd.Prim] = spawn_particles_pyramid

    # Distribution
    dim_x: int = MISSING  # type: ignore
    dim_y: int = MISSING  # type: ignore
    dim_z: int = MISSING  # type: ignore

    # Initial velocity
    velocity: Tuple[
        float | Tuple[float, float],
        float | Tuple[float, float],
        float | Tuple[float, float],
    ] = (0.0, 0.0, 0.0)



from __future__ import annotations

from typing import Sequence, Type

from srb.core.asset.robot.mobile.aerial_robot import AerialRobot
from srb.core.asset.robot.mobile_manipulation.mobile_manipulator import (
    MobileManipulatorRegistry,
)
from srb.core.asset.robot.mobile_manipulation.mobile_manipulator_type import (
    MobileManipulatorType,
)

from ._metaclass import CombinedMobileManipulator


class AerialManipulator(
    CombinedMobileManipulator,
    AerialRobot,
    mobile_manipulator_entrypoint=MobileManipulatorType.AERIAL,
):
    ## Model
    mobile_base: AerialRobot | None = None

    @classmethod
    def mobile_manipulator_registry(cls) -> Sequence[Type[AerialManipulator]]:
        return MobileManipulatorRegistry.registry.get(MobileManipulatorType.AERIAL, [])  # type: ignore



import importlib
import pkgutil
import sys
from typing import Iterable, List

from srb.utils import logging


def import_recursively(module_name: str, ignorelist: List[str] = []):
    package = importlib.import_module(module_name)
    for _ in _import_recursively_impl(
        path=package.__path__, prefix=f"{package.__name__}.", ignorelist=ignorelist
    ):
        pass


def _import_recursively_impl(
    path: Iterable[str],
    prefix: str = "",
    ignorelist: List[str] = [],
) -> Iterable[pkgutil.ModuleInfo]:
    def seen(p, m={}):
        if p in m:
            return True
        m[p] = True

    for info in pkgutil.iter_modules(path, prefix):
        if any(module_name in info.name for module_name in ignorelist):
            continue

        yield info

        if info.ispkg:
            try:
                __import__(info.name)
            except Exception as e:
                logging.critical(f"Failed to import '{info.name}'")
                raise e
            else:
                paths = getattr(sys.modules[info.name], "__path__", None) or []
                paths = [path for path in paths if not seen(path)]
                yield from _import_recursively_impl(paths, f"{info.name}.", ignorelist)



from .light import *  # noqa: F403
from .payload import *  # noqa: F403
from .pedestal import *  # noqa: F403
from .tool import *  # noqa: F403

# isort: split

from .asteroid import Asteroid  # noqa: F401
from .beneficiation_unit import BeneficiationUnit  # noqa: F401
from .bolt_and_nut import BoltM8, NutM8  # noqa: F401
from .juggling_ball import JugglingBall  # noqa: F401
from .peg_in_hole import (  # noqa: F401
    Hole,
    Peg,
    ProfileHole,
    ProfilePeg,
    ShortProfilePeg,
)
from .rock import LunarRock, MarsRock  # noqa: F401
from .sample import SampleTube  # noqa: F401
from .shape import RandomShape  # noqa: F401
from .solar_panel import SolarPanel  # noqa: F401



from srb.core.action import (  # noqa: F401
    ActionGroup,
    DifferentialIKControllerCfg,
    DifferentialInverseKinematicsActionCfg,
    InverseKinematicsActionGroup,
    JointEffortActionCfg,
    JointEffortActionGroup,
    JointPositionRelativeActionGroup,
    OperationalSpaceControlActionGroup,
    OperationalSpaceControllerActionCfg,
    OperationalSpaceControllerCfg,
    RelativeJointPositionActionCfg,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, SerialManipulator, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    MeshCollisionPropertiesCfg,
    MultiAssetSpawnerCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import deg_to_rad, rpy_to_quat
from srb.utils.nucleus import ISAAC_NUCLEUS_DIR, ISAACLAB_NUCLEUS_DIR


class UR3(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/ur3",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/UniversalRobots/ur3/ur3.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "shoulder_pan_joint": 0.0,
                "shoulder_lift_joint": deg_to_rad(-90.0),
                "elbow_joint": deg_to_rad(90.0),
                "wrist_1_joint": deg_to_rad(-90.0),
                "wrist_2_joint": deg_to_rad(-90.0),
                "wrist_3_joint": deg_to_rad(-90.0),
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=100.0,
                effort_limit=87.0,
                stiffness=800.0,
                damping=40.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*_joint"],
            base_name="base_link",
            body_name="wrist_3_link",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="wrist_3_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base_link")
    frame_flange: Frame = Frame(prim_relpath="wrist_3_link")
    frame_base_camera: Frame = Frame(
        prim_relpath="base_link/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="wrist_3_link/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )


class UR3e(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/ur3e",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/UniversalRobots/ur3e/ur3e.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "shoulder_pan_joint": 0.0,
                "shoulder_lift_joint": deg_to_rad(-90.0),
                "elbow_joint": deg_to_rad(90.0),
                "wrist_1_joint": deg_to_rad(-90.0),
                "wrist_2_joint": deg_to_rad(-90.0),
                "wrist_3_joint": deg_to_rad(-90.0),
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=100.0,
                effort_limit=87.0,
                stiffness=800.0,
                damping=40.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*_joint"],
            base_name="base_link",
            body_name="wrist_3_link",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="wrist_3_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base_link")
    frame_flange: Frame = Frame(prim_relpath="wrist_3_link")
    frame_base_camera: Frame = Frame(
        prim_relpath="base_link/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="wrist_3_link/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )


class UR5(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/ur5",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/UniversalRobots/ur5/ur5.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "shoulder_pan_joint": 0.0,
                "shoulder_lift_joint": deg_to_rad(-90.0),
                "elbow_joint": deg_to_rad(90.0),
                "wrist_1_joint": deg_to_rad(-90.0),
                "wrist_2_joint": deg_to_rad(-90.0),
                "wrist_3_joint": deg_to_rad(-90.0),
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=100.0,
                effort_limit=87.0,
                stiffness=800.0,
                damping=40.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*_joint"],
            base_name="base_link",
            body_name="wrist_3_link",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="wrist_3_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base_link")
    frame_flange: Frame = Frame(prim_relpath="wrist_3_link")
    frame_base_camera: Frame = Frame(
        prim_relpath="base_link/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="wrist_3_link/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )


class UR5e(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/ur5e",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/UniversalRobots/ur5e/ur5e.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "shoulder_pan_joint": 0.0,
                "shoulder_lift_joint": deg_to_rad(-90.0),
                "elbow_joint": deg_to_rad(90.0),
                "wrist_1_joint": deg_to_rad(-90.0),
                "wrist_2_joint": deg_to_rad(-90.0),
                "wrist_3_joint": deg_to_rad(-90.0),
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=100.0,
                effort_limit=87.0,
                stiffness=800.0,
                damping=40.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*_joint"],
            base_name="base_link",
            body_name="wrist_3_link",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="wrist_3_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="fixed",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_task=100.0,
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="wrist_3_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="wrist_3_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_limits_task=(0.5, 2.5),
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         damping_ratio_scale=1.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base_link")
    frame_flange: Frame = Frame(prim_relpath="wrist_3_link")
    frame_base_camera: Frame = Frame(
        prim_relpath="base_link/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="wrist_3_link/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )


class UR10(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/ur10",
        spawn=UsdFileCfg(
            usd_path=f"{ISAACLAB_NUCLEUS_DIR}/Robots/UniversalRobots/UR10/ur10_instanceable.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "shoulder_pan_joint": 0.0,
                "shoulder_lift_joint": deg_to_rad(-90.0),
                "elbow_joint": deg_to_rad(90.0),
                "wrist_1_joint": deg_to_rad(-90.0),
                "wrist_2_joint": deg_to_rad(-90.0),
                "wrist_3_joint": deg_to_rad(-90.0),
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=100.0,
                effort_limit=87.0,
                stiffness=800.0,
                damping=40.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*_joint"],
            base_name="base_link",
            body_name="ee_link",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="ee_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base_link")
    frame_flange: Frame = Frame(
        prim_relpath="ee_link",
        offset=Transform(rot=rpy_to_quat(180.0, -90.0, 0.0)),
    )
    frame_base_camera: Frame = Frame(
        prim_relpath="base_link/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="ee_link/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )


class UR10e(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/ur10e",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/UniversalRobots/ur10e/ur10e.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "shoulder_pan_joint": 0.0,
                "shoulder_lift_joint": deg_to_rad(-90.0),
                "elbow_joint": deg_to_rad(90.0),
                "wrist_1_joint": deg_to_rad(-90.0),
                "wrist_2_joint": deg_to_rad(-90.0),
                "wrist_3_joint": deg_to_rad(-90.0),
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=100.0,
                effort_limit=87.0,
                stiffness=800.0,
                damping=40.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*_joint"],
            base_name="base_link",
            body_name="wrist_3_link",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="wrist_3_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base_link")
    frame_flange: Frame = Frame(prim_relpath="wrist_3_link")
    frame_base_camera: Frame = Frame(
        prim_relpath="base_link/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="wrist_3_link/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )


class UR16e(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/ur16e",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/UniversalRobots/ur16e/ur16e.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "shoulder_pan_joint": 0.0,
                "shoulder_lift_joint": deg_to_rad(-90.0),
                "elbow_joint": deg_to_rad(90.0),
                "wrist_1_joint": deg_to_rad(-90.0),
                "wrist_2_joint": deg_to_rad(-90.0),
                "wrist_3_joint": deg_to_rad(-90.0),
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=100.0,
                effort_limit=87.0,
                stiffness=800.0,
                damping=40.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*_joint"],
            base_name="base_link",
            body_name="wrist_3_link",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="wrist_3_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base_link")
    frame_flange: Frame = Frame(prim_relpath="wrist_3_link")
    frame_base_camera: Frame = Frame(
        prim_relpath="base_link/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="wrist_3_link/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )


class UR20(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/ur20",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/UniversalRobots/ur20/ur20.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "shoulder_pan_joint": 0.0,
                "shoulder_lift_joint": deg_to_rad(-90.0),
                "elbow_joint": deg_to_rad(90.0),
                "wrist_1_joint": deg_to_rad(-90.0),
                "wrist_2_joint": deg_to_rad(-90.0),
                "wrist_3_joint": deg_to_rad(-90.0),
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=100.0,
                effort_limit=87.0,
                stiffness=800.0,
                damping=40.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*_joint"],
            base_name="base_link",
            body_name="wrist_3_link",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="wrist_3_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base_link")
    frame_flange: Frame = Frame(prim_relpath="wrist_3_link")
    frame_base_camera: Frame = Frame(
        prim_relpath="base_link/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="wrist_3_link/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )


class UR30(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/ur30",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/UniversalRobots/ur30/ur30.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "shoulder_pan_joint": 0.0,
                "shoulder_lift_joint": deg_to_rad(-90.0),
                "elbow_joint": deg_to_rad(90.0),
                "wrist_1_joint": deg_to_rad(-90.0),
                "wrist_2_joint": deg_to_rad(-90.0),
                "wrist_3_joint": deg_to_rad(-90.0),
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=100.0,
                effort_limit=87.0,
                stiffness=800.0,
                damping=40.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*_joint"],
            base_name="base_link",
            body_name="wrist_3_link",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="wrist_3_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base_link")
    frame_flange: Frame = Frame(prim_relpath="wrist_3_link")
    frame_base_camera: Frame = Frame(
        prim_relpath="base_link/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="wrist_3_link/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )


class RandomURManipulator(SerialManipulator):
    asset_cfg: ArticulationCfg = UR30().asset_cfg.copy()  # type: ignore
    asset_cfg.prim_path = "{ENV_REGEX_NS}/anymal"
    asset_cfg.spawn = MultiAssetSpawnerCfg(
        random_choice=False,
        assets_cfg=(
            UR3e().asset_cfg.spawn,  # type: ignore
            UR5().asset_cfg.spawn,  # type: ignore
            UR10e().asset_cfg.spawn,  # type: ignore
            UR16e().asset_cfg.spawn,  # type: ignore
            UR20().asset_cfg.spawn,  # type: ignore
            UR30().asset_cfg.spawn,  # type: ignore
        ),
        activate_contact_sensors=True,
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*_joint"],
            base_name="base_link",
            body_name="wrist_3_link",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*_joint"],
    #         body_name="wrist_3_link",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base_link")
    frame_flange: Frame = Frame(prim_relpath="wrist_3_link")
    frame_base_camera: Frame = Frame(
        prim_relpath="base_link/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="wrist_3_link/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )



from dataclasses import MISSING

from srb import assets
from srb.core.asset import (
    Articulation,
    AssetBaseCfg,
    AssetVariant,
    Manipulator,
    MobileRobot,
    Pedestal,
    RigidObject,
)
from srb.core.env import BaseEventCfg, BaseSceneCfg, DirectEnv, DirectEnvCfg, ViewerCfg
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.marker import FRAME_MARKER_SMALL_CFG
from srb.core.mdp import reset_joints_by_offset
from srb.core.sensor import (
    ContactSensor,
    ContactSensorCfg,
    FrameTransformer,
    FrameTransformerCfg,
)
from srb.utils.cfg import configclass
from srb.utils.math import combine_frame_transforms_tuple, deg_to_rad


@configclass
class ManipulationSceneCfg(BaseSceneCfg):
    env_spacing: float = 4.0

    ## Assets
    pedestal: AssetBaseCfg | None = None

    ## Sensors
    tf_end_effector: FrameTransformerCfg = FrameTransformerCfg(
        prim_path=MISSING,  # type: ignore
        target_frames=[
            FrameTransformerCfg.FrameCfg(
                name="robot_ee",
                prim_path=MISSING,  # type: ignore
            ),
        ],
        visualizer_cfg=FRAME_MARKER_SMALL_CFG.replace(prim_path="/Visuals/robot_ee"),
    )
    contacts_robot: ContactSensorCfg = ContactSensorCfg(
        prim_path=MISSING,  # type: ignore
    )
    contacts_end_effector: ContactSensorCfg | None = None


@configclass
class ManipulationEventCfg(BaseEventCfg):
    randomize_robot_joints: EventTermCfg = EventTermCfg(
        func=reset_joints_by_offset,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("robot"),
            "position_range": (-deg_to_rad(5.0), deg_to_rad(5.0)),
            "velocity_range": (0.0, 0.0),
        },
    )


@configclass
class ManipulationEnvCfg(DirectEnvCfg):
    ## Assets
    robot: Manipulator | AssetVariant = assets.Franka()
    _robot: Manipulator = MISSING  # type: ignore
    pedestal: Pedestal | MobileRobot | None = assets.IndustrialPedestal25()

    ## Scene
    scene: ManipulationSceneCfg = ManipulationSceneCfg()

    ## Events
    events: ManipulationEventCfg = ManipulationEventCfg()

    ## Time
    env_rate: float = 1.0 / 150.0
    agent_rate: float = 1.0 / 50.0

    ## Viewer
    viewer: ViewerCfg = ViewerCfg(
        eye=(1.85, 0.0, 1.85), lookat=(0.125, 0.0, 0.25), origin_type="env"
    )

    def __post_init__(self):
        super().__post_init__()

        ## Add pedestal and offset the robot
        if self.pedestal is not None:
            if isinstance(self.pedestal.asset_cfg, AssetBaseCfg):
                self.scene.pedestal = self.pedestal.asset_cfg
            else:
                self.scene.pedestal = AssetBaseCfg(
                    prim_path=self.pedestal.asset_cfg.prim_path,
                    spawn=self.pedestal.asset_cfg.spawn,
                    init_state=self.pedestal.asset_cfg.init_state,
                )
                for props in (
                    "articulation_props",
                    "deformable_props",
                    "fixed_tendons_props",
                    "joint_drive_props",
                    "mass_props",
                    "rigid_props",
                ):
                    setattr(self.scene.pedestal.spawn, props, None)
            self.scene.robot.init_state.pos, self.scene.robot.init_state.rot = (
                combine_frame_transforms_tuple(
                    self._robot.asset_cfg.init_state.pos,
                    self._robot.asset_cfg.init_state.rot,
                    self.pedestal.frame_manipulator_mount.offset.pos,
                    self.pedestal.frame_manipulator_mount.offset.rot,
                )
            )
            self.scene.pedestal.prim_path = (
                "/World/pedestal" if self.stack else "{ENV_REGEX_NS}/pedestal"
            )

        # Sensor: End-effector transform
        self.scene.tf_end_effector.prim_path = (
            f"{self.scene.robot.prim_path}/{self._robot.frame_base.prim_relpath}"
        )
        self.scene.tf_end_effector.target_frames[
            0
        ].prim_path = (
            f"{self.scene.robot.prim_path}/{self._robot.frame_flange.prim_relpath}"
        )
        if self._robot.end_effector is not None:
            (
                self.scene.tf_end_effector.target_frames[0].offset.pos,
                self.scene.tf_end_effector.target_frames[0].offset.rot,
            ) = combine_frame_transforms_tuple(
                self._robot.frame_flange.offset.pos,
                self._robot.frame_flange.offset.rot,
                self._robot.end_effector.frame_tool_centre_point.offset.pos,
                self._robot.end_effector.frame_tool_centre_point.offset.rot,
            )
        else:
            (
                self.scene.tf_end_effector.target_frames[0].offset.pos,
                self.scene.tf_end_effector.target_frames[0].offset.rot,
            ) = (
                self._robot.frame_flange.offset.pos,
                self._robot.frame_flange.offset.rot,
            )

        # Sensor: Robot contacts
        self.scene.contacts_robot.prim_path = f"{self.scene.robot.prim_path}/.*"

        # Sensor: End-effector contacts
        self.scene.contacts_end_effector = (
            ContactSensorCfg(
                prim_path=f"{self._robot.end_effector.asset_cfg.prim_path}/.*",
            )
            if self._robot.end_effector is not None
            else None
        )


class ManipulationEnv(DirectEnv):
    cfg: ManipulationEnvCfg

    def __init__(self, cfg: ManipulationEnvCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._tf_end_effector: FrameTransformer = self.scene["tf_end_effector"]
        self._contacts_robot: ContactSensor = self.scene["contacts_robot"]
        self._end_effector: Articulation | RigidObject | None = (
            self.scene.articulations.get("end_effector", None)
            or self.scene.rigid_objects.get("end_effector", None)
            or None
        )
        self._contacts_end_effector: ContactSensor | None = self.scene.sensors.get(  # type: ignore
            "contacts_end_effector", None
        )



from dataclasses import MISSING

import torch

from srb.core.action.action_group import ActionGroup
from srb.core.action.term import MulticopterBodyAccelerationActionCfg
from srb.utils.cfg import configclass


@configclass
class MulticopterBodyAccelerationActionGroup(ActionGroup):
    cmd_vel: MulticopterBodyAccelerationActionCfg = MISSING  # type: ignore

    def map_cmd_to_action(self, twist: torch.Tensor, event: bool) -> torch.Tensor:
        return torch.concat(
            (
                twist[:3],
                twist[5].unsqueeze(0),
            ),
        )



from __future__ import annotations

from functools import cached_property
from typing import ClassVar, Dict, Iterable, List, Sequence, Tuple, Type

from pydantic import PositiveFloat

from srb.core.asset import AssetBaseCfg
from srb.core.asset.asset import Asset, AssetRegistry
from srb.core.asset.asset_type import AssetType
from srb.core.asset.scenery.scenery_type import SceneryType


class Scenery(Asset, asset_entrypoint=AssetType.SCENERY):
    ## Model
    asset_cfg: AssetBaseCfg

    ## Attributes forwarded to the spawner
    SPAWNER_ATTRIBUTES: ClassVar[Sequence[str]] = ("density", "flat_area_size")
    density: PositiveFloat | None = None
    flat_area_size: PositiveFloat | None = None

    def __init_subclass__(
        cls,
        scenery_entrypoint: SceneryType | None = None,
        scenery_metaclass: bool = False,
        asset_metaclass: bool = False,
        **kwargs,
    ):
        super().__init_subclass__(
            asset_metaclass=(
                asset_metaclass or scenery_entrypoint is not None or scenery_metaclass
            ),
            **kwargs,
        )
        if scenery_entrypoint is not None:
            assert isinstance(scenery_entrypoint, SceneryType), (
                f"Class '{cls.__name__}' is marked as a scenery entrypoint, but '{scenery_entrypoint}' is not a valid {SceneryType}"
            )
            assert scenery_entrypoint not in SceneryRegistry.base_types.keys(), (
                f"Class '{cls.__name__}' is marked as '{scenery_entrypoint}' scenery entrypoint, but it was already marked by '{SceneryRegistry.base_types[scenery_entrypoint].__name__}'"
            )
            SceneryRegistry.base_types[scenery_entrypoint] = cls
        elif scenery_metaclass:
            SceneryRegistry.meta_types.append(cls)
        elif not asset_metaclass:
            for scenery_type, base in SceneryRegistry.base_types.items():
                if issubclass(cls, base):
                    if scenery_type not in SceneryRegistry.registry.keys():
                        SceneryRegistry.registry[scenery_type] = []
                    else:
                        assert cls.name() not in (
                            scenery.name()
                            for scenery in SceneryRegistry.registry[scenery_type]
                        ), (
                            f"Cannot register multiple sceneries with an identical name: '{cls.__module__}:{cls.__name__}' already exists as '{next(scenery for scenery in SceneryRegistry.registry[scenery_type] if cls.name() == scenery.name()).__module__}:{cls.__name__}'"
                        )
                    SceneryRegistry.registry[scenery_type].append(cls)
                    break

    @cached_property
    def scenery_type(self) -> SceneryType:
        for scenery_type, base in SceneryRegistry.base_types.items():
            if isinstance(self, base):
                return scenery_type
        raise ValueError(f"Class '{self.__class__.__name__}' has unknown scenery type")

    @classmethod
    def scenery_registry(cls) -> Sequence[Type[Scenery]]:
        return list(SceneryRegistry.values_inner())

    @classmethod
    def asset_registry(cls) -> Sequence[Type[Scenery]]:
        return AssetRegistry.registry.get(AssetType.SCENERY, [])  # type: ignore


class SceneryRegistry:
    registry: ClassVar[Dict[SceneryType, List[Type[Scenery]]]] = {}
    base_types: ClassVar[Dict[SceneryType, Type[Scenery]]] = {}
    meta_types: ClassVar[List[Type[Scenery]]] = []

    @classmethod
    def keys(cls) -> Iterable[SceneryType]:
        return cls.registry.keys()

    @classmethod
    def items(cls) -> Iterable[Tuple[SceneryType, Sequence[Type[Scenery]]]]:
        return cls.registry.items()

    @classmethod
    def values(cls) -> Iterable[Iterable[Type[Scenery]]]:
        return cls.registry.values()

    @classmethod
    def values_inner(cls) -> Iterable[Type[Scenery]]:
        return (scenery for sceneries in cls.registry.values() for scenery in sceneries)

    @classmethod
    def n_sceneries(cls) -> int:
        return sum(len(sceneries) for sceneries in cls.registry.values())

    @classmethod
    def registered_modules(cls) -> Iterable[str]:
        return {
            scenery.__module__
            for sceneries in cls.registry.values()
            for scenery in sceneries
        }

    @classmethod
    def registered_packages(cls) -> Iterable[str]:
        return {module.split(".", maxsplit=1)[0] for module in cls.registered_modules()}

    @classmethod
    def get_by_name(cls, name: str) -> Type[Scenery] | None:
        for scenery in cls.values_inner():
            if scenery.name() == name:
                return scenery
        return None



from isaaclab.envs.common import (  # noqa: F401
    ActionType,
    AgentID,
    EnvStepReturn,
    ObsType,
    SpaceType,
    StateType,
    VecEnvObs,
    VecEnvStepReturn,
)

from .cfg import AnyEnv, AnyEnvCfg  # noqa: F401
from .enum import EnumNameSerializer  # noqa: F401
from .step_return import StepReturn, _StepReturn  # noqa: F401



from __future__ import annotations

from functools import cache
from typing import ClassVar, Iterable, List, Sequence, Type

import torch

from srb.utils.cfg import configclass
from srb.utils.str import convert_to_snake_case


@configclass
class ActionGroup:
    @classmethod
    @cache
    def name(cls) -> str:
        return convert_to_snake_case(cls.__name__).removesuffix("_action_group")

    def map_cmd_to_action(self, twist: torch.Tensor, event: bool) -> torch.Tensor:
        raise NotImplementedError()

    def __init_subclass__(cls, action_group_metaclass: bool = False, **kwargs):
        super().__init_subclass__(**kwargs)
        if action_group_metaclass:
            return
        assert cls.name() not in (
            action_group.name() for action_group in ActionGroupRegistry.registry
        ), (
            f"Cannot register multiple action groups with an identical name: '{cls.__module__}:{cls.__name__}' already exists as '{next(robot for robot in ActionGroupRegistry.registry if cls.name() == robot.name()).__module__}:{cls.__name__}'"
        )
        ActionGroupRegistry.registry.append(cls)

    @classmethod
    def action_group_registry(cls) -> Sequence[Type[ActionGroup]]:
        return ActionGroupRegistry.registry


class ActionGroupRegistry:
    registry: ClassVar[List[Type[ActionGroup]]] = []

    @classmethod
    def __len__(cls) -> int:
        return len(cls.registry)

    @classmethod
    def registered_modules(cls) -> Iterable[str]:
        return {action_group.__module__ for action_group in cls.registry}

    @classmethod
    def registered_packages(cls) -> Iterable[str]:
        return {module.split(".", maxsplit=1)[0] for module in cls.registered_modules()}

    @classmethod
    def get_by_name(cls, name: str) -> Type[ActionGroup] | None:
        for action_group in cls.registry:
            if action_group.name() == name:
                return action_group
        return None



from dataclasses import MISSING

from srb.core.asset import ArticulationCfg, AssetBaseCfg, RigidObjectCfg
from srb.core.env import InteractiveSceneCfg
from srb.utils.cfg import configclass


@configclass
class BaseSceneCfg(InteractiveSceneCfg):
    num_envs: int = 1
    replicate_physics: bool = False

    ## Illumination
    sunlight: AssetBaseCfg | None = None
    skydome: AssetBaseCfg | None = None

    ## Scenery
    scenery: AssetBaseCfg | None = None

    ## Robot
    robot: RigidObjectCfg | ArticulationCfg = MISSING  # type: ignore
    payload: AssetBaseCfg | None = None
    manipulator: ArticulationCfg | None = None
    end_effector: AssetBaseCfg | None = None



from .env import (  # noqa: F401
    AerialManipulationEnv,
    AerialManipulationEnvCfg,
    AerialManipulationEventCfg,
    AerialManipulationSceneCfg,
)
from .visual_ext import AerialManipulationEnvVisualExtCfg  # noqa: F401



from dataclasses import MISSING

import torch

from srb.core.action import (
    DifferentialInverseKinematicsActionCfg,
    OperationalSpaceControllerActionCfg,
)
from srb.core.action.action_group import ActionGroup
from srb.utils.cfg import configclass


@configclass
class InverseKinematicsActionGroup(ActionGroup):
    delta_twist: DifferentialInverseKinematicsActionCfg = MISSING  # type: ignore

    def map_cmd_to_action(self, twist: torch.Tensor, event: bool) -> torch.Tensor:
        if self.delta_twist.controller.command_type == "pose":
            assert self.delta_twist.controller.use_relative_mode, (
                "Only relative mode is supported"
            )
            return twist
        else:
            return twist[:3]


@configclass
class OperationalSpaceControlActionGroup(ActionGroup):
    delta_twist: OperationalSpaceControllerActionCfg = MISSING  # type: ignore

    def map_cmd_to_action(self, twist: torch.Tensor, event: bool) -> torch.Tensor:
        assert (
            len(self.delta_twist.controller_cfg.target_types) == 1
            and "pose_rel" in self.delta_twist.controller_cfg.target_types
        ), "Only pose_rel OSC target type is supported for teleoperation"
        match self.delta_twist.controller_cfg.impedance_mode:
            case "fixed":
                return twist
            case "variable_kp":
                return torch.cat(
                    (
                        twist,
                        torch.ones(6, dtype=twist.dtype, device=twist.device),
                    ),
                    dim=-1,
                )
            case "variable":
                return torch.cat(
                    (
                        twist,
                        torch.ones(12, dtype=twist.dtype, device=twist.device),
                    ),
                    dim=-1,
                )
            case _:
                raise ValueError(
                    f"Unknown impedance mode: {self.delta_twist.controller_cfg.impedance_mode}"
                )



from enum import Enum, auto
from typing import TYPE_CHECKING, List, Optional, Sequence, Tuple, Type

import torch
from pydantic import BaseModel

from srb.core.manager import ActionTerm, ActionTermCfg
from srb.utils import logging
from srb.utils.cfg import configclass
from srb.utils.math import deg_to_rad

if TYPE_CHECKING:
    from srb._typing import AnyEnv
    from srb.core.asset import Articulation


class WheeledDriveAction(ActionTerm):
    cfg: "WheeledDriveActionCfg"
    _env: "AnyEnv"
    _asset: "Articulation"

    def __init__(self, cfg: "WheeledDriveActionCfg", env: "AnyEnv"):
        super().__init__(
            cfg,
            env,  # type: ignore
        )

        ## Auto-detect drive type if not specified
        if self.cfg.drive_type is None:
            self.drive_type = self._autodetect_drive_type()
            logging.debug(f"Auto-detected drive type: {self.drive_type}")
        else:
            self.drive_type = self.cfg.drive_type

        ## Prepare wheel configuration
        self._setup_wheel_configuration()

    @property
    def action_dim(self) -> int:
        return 2  # linear_velocity, angular_velocity

    @property
    def raw_actions(self) -> torch.Tensor:
        return self._raw_actions

    @property
    def processed_actions(self) -> torch.Tensor:
        return self._processed_actions

    def process_actions(self, actions: torch.Tensor):
        self._raw_actions = actions
        self._processed_actions = actions.clone()
        self._processed_actions[:, 0:3] *= self.cfg.scale_linear
        self._processed_actions[:, 3:6] *= self.cfg.scale_linear
        if self.cfg.max_linear_velocity is not None:
            self._processed_actions[:, 0].clamp_(
                -self.cfg.max_linear_velocity, self.cfg.max_linear_velocity
            )
        if self.cfg.max_angular_velocity is not None:
            self._processed_actions[:, 1].clamp_(
                -self.cfg.max_angular_velocity, self.cfg.max_angular_velocity
            )

    def apply_actions(self):
        ## Calculate drive commands based on drive type
        match self.drive_type:
            case DriveType.SKID_STEER:
                drive_velocities, steer_angles = self._skid_steer_drive(
                    self._processed_actions[:, 0], self._processed_actions[:, 1]
                )
            case DriveType.ACKERMANN:
                drive_velocities, steer_angles = self._ackermann_drive(
                    self._processed_actions[:, 0], self._processed_actions[:, 1]
                )
            case DriveType.ROVER:
                drive_velocities, steer_angles = self._rover_drive(
                    self._processed_actions[:, 0], self._processed_actions[:, 1]
                )

        ## Apply drive commands
        self._asset.set_joint_velocity_target(
            drive_velocities, joint_ids=self._drive_joint_indices
        )

        ## Apply steering commands
        if steer_angles is not None:
            self._asset.set_joint_position_target(
                steer_angles, joint_ids=self._steering_joint_indices
            )

    def reset(self, env_ids: Sequence[int] | None = None):
        super().reset(env_ids)

    def _skid_steer_drive(
        self, v_lin: torch.Tensor, v_ang: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor | None]:
        drive_velocities = torch.zeros(
            (self.num_envs, len(self._drive_wheels)), device=self.device
        )
        for i, wheel_idx in enumerate(self._drive_wheels):
            y_offset = self._wheel_positions[wheel_idx, 1]
            drive_velocities[:, i] = (v_lin + v_ang * y_offset) / self.cfg.wheel_radius

        return drive_velocities, None

    def _ackermann_drive(
        self, v_lin: torch.Tensor, v_ang: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor | None]:
        drive_velocities = torch.zeros(
            (self.num_envs, len(self._drive_wheels)), device=self.device
        )
        steer_angles = torch.zeros(
            (self.num_envs, len(self._steering_wheels)), device=self.device
        )

        ## Turning radius and direction
        turn_radius = torch.where(
            v_ang != 0.0,
            torch.abs(v_lin) / torch.abs(v_ang),
            torch.ones_like(v_lin) * 1.0e10,
        )
        turn_dir = torch.sign(v_ang)

        # Wheelbase parameters
        wheelbase_length = torch.tensor(self.cfg.wheelbase[0], device=self.device)

        # Calculate steering angles for front wheels
        for i, wheel_idx in enumerate(self._steering_wheels):
            wheel_pos = self._wheel_positions[wheel_idx]
            # Ackermann geometry
            effective_radius = turn_radius - turn_dir * wheel_pos[1]
            angle = turn_dir * torch.atan2(wheelbase_length, effective_radius)
            # Apply steering limits
            angle = torch.clamp(
                angle, -self._steering_limits[i], self._steering_limits[i]
            )
            steer_angles[:, i] = angle

        # Calculate wheel velocities
        for i, wheel_idx in enumerate(self._drive_wheels):
            wheel_y = self._wheel_positions[wheel_idx, 1]
            # Path radius for this wheel
            wheel_path_radius = turn_radius - turn_dir * wheel_y
            # Velocity calculation
            drive_velocities[:, i] = (
                torch.where(
                    v_ang != 0,
                    torch.abs(v_ang) * wheel_path_radius * torch.sign(v_lin),
                    v_lin,
                )
                / self.cfg.wheel_radius
            )

        return drive_velocities, steer_angles

    def _rover_drive(
        self, v_lin: torch.Tensor, v_ang: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor | None]:
        drive_velocities = torch.zeros(
            (self.num_envs, len(self._drive_wheels)), device=self.device
        )
        steer_angles = torch.zeros(
            (self.num_envs, len(self._steering_wheels)), device=self.device
        )

        # Key parameters
        abs_v_lin = torch.abs(v_lin)
        abs_v_ang = torch.abs(v_ang)
        turn_dir = torch.sign(v_ang)
        drive_dir = torch.sign(v_lin)
        drive_dir = torch.where(drive_dir == 0, torch.ones_like(drive_dir), drive_dir)

        # Wheelbase parameters
        half_width = torch.tensor(self.cfg.wheelbase[1] / 2, device=self.device)

        # Calculate turning radius
        turning_radius = torch.where(
            abs_v_ang > 0,
            abs_v_lin / abs_v_ang,
            torch.tensor(1.0e10, device=self.device),
        )

        # Detect point turn mode (tight radius)
        is_point_turn = turning_radius == 0.0

        # Calculate wheel path radii
        r_left = turning_radius - half_width * turn_dir
        r_right = turning_radius + half_width * turn_dir

        # Calculate steering angles for corner wheels
        for i, idx in enumerate(self._steering_wheels):
            wheel_pos = self._wheel_positions[idx]
            is_front = wheel_pos[0] > 0
            is_left = wheel_pos[1] < 0

            # Point turn: wheels turn inward/outward
            point_angle_val = 0.78539816 * (-1 if (is_front == is_left) else 1)

            # Normal driving: Reduce steering angle for rover mode
            # Calculate the base angle but scale it down to prevent extreme angles
            base_angle = torch.atan2(half_width, r_left if is_left else r_right)
            # Apply scaling factor to reduce steering angle (0.3 = ~30% of the original angle)
            scaled_angle = base_angle * 0.3

            # Front wheels turn in direction of turn, rear wheels in opposite direction
            # This is critical for proper rover steering geometry
            normal_angle = turn_dir * scaled_angle * (1 if is_front else -1)

            # Select angle based on mode
            steer_angles[:, i] = torch.where(
                is_point_turn,
                point_angle_val,
                normal_angle,
            )

        # Calculate wheel velocities
        for i, wheel_idx in enumerate(self._drive_wheels):
            wheel_pos = self._wheel_positions[wheel_idx]
            is_left = wheel_pos[1] < 0

            # Directional factor for point turns
            side_factor = -1.0 if is_left else 1.0

            # Velocity profiles for different modes
            point_turn_vel = side_factor * turn_dir * abs_v_ang
            normal_vel = drive_dir * torch.where(
                abs_v_ang == 0,
                abs_v_lin,  # Straight motion
                abs_v_ang,  # Normal turning
            )

            # Final wheel velocity
            wheel_vel = torch.where(is_point_turn, point_turn_vel, normal_vel)
            drive_velocities[:, i] = wheel_vel / self.cfg.wheel_radius

        return drive_velocities, steer_angles

    def _autodetect_drive_type(self) -> "DriveType":
        match len(self.cfg.steering_joint_names):
            case 0:
                return DriveType.SKID_STEER
            case _len if _len < 4:
                return DriveType.ACKERMANN
            case _:
                return DriveType.ROVER

    def _setup_wheel_configuration(self):
        # Get joint indices
        assert len(self.cfg.drive_joint_names) > 0, "No drive joints specified"
        self._drive_joint_indices = self._asset.find_joints(
            self.cfg.drive_joint_names, preserve_order=True
        )[0]
        self._steering_joint_indices = (
            self._asset.find_joints(self.cfg.steering_joint_names, preserve_order=True)[
                0
            ]
            if self.cfg.steering_joint_names
            else None
        )

        # Determine number of wheels from drive joints
        num_wheels = len(self._drive_joint_indices)

        # Generate positions based on wheelbase dimensions
        positions = self._generate_wheel_positions(num_wheels)
        logging.debug(f"Generated wheel positions for {num_wheels} wheels")

        # Store wheel positions as tensor
        self._wheel_positions = torch.tensor(positions, device=self.device)

        # All wheels are drive wheels
        self._drive_wheels = list(range(num_wheels))

        # Determine which wheels have steering capability
        if not self.cfg.steering_joint_names:
            self._steering_wheels = []
        else:
            # Map steering joints to corresponding drive wheels
            self._steering_wheels = self._map_steering_to_drive_wheels(positions)

        # Set steering limits based on drive type
        # More angle for point turns, less for normal driving
        if self.drive_type == DriveType.ROVER:
            steering_limit = deg_to_rad(45)
        else:
            steering_limit = deg_to_rad(35)

        self._steering_limits = (
            torch.ones(len(self._steering_wheels), device=self.device) * steering_limit
        )

        # Set a separate limit for normal driving in rover mode (used in _rover_drive)
        if self.drive_type == DriveType.ROVER:
            self._normal_steering_limit = deg_to_rad(22.5)

    def _generate_wheel_positions(self, num_wheels):
        half_length = self.cfg.wheelbase[0] / 2
        half_width = self.cfg.wheelbase[1] / 2

        if num_wheels == 2:
            # Simple 2-wheel differential drive
            return [
                (0, -half_width, 0),  # Left
                (0, half_width, 0),  # Right
            ]
        elif num_wheels == 4:
            # 4-wheel layout (common car-like or differential)
            return [
                (half_length, -half_width, 0),  # Front left
                (half_length, half_width, 0),  # Front right
                (-half_length, -half_width, 0),  # Rear left
                (-half_length, half_width, 0),  # Rear right
            ]
        elif num_wheels == 6:
            # 6-wheel rover layout
            mid_width = (self.cfg.wheelbase_mid or self.cfg.wheelbase[1]) / 2
            return [
                (half_length, -half_width, 0),  # Front left
                (half_length, half_width, 0),  # Front right
                (0, -mid_width, 0),  # Middle left
                (0, mid_width, 0),  # Middle right
                (-half_length, -half_width, 0),  # Rear left
                (-half_length, half_width, 0),  # Rear right
            ]
        else:
            # For other wheel counts, distribute evenly
            positions = []
            if num_wheels % 2 == 0:
                # Even number of wheels - arrange symmetrically in pairs
                pairs = num_wheels // 2
                for i in range(pairs):
                    x = half_length - i * (2 * half_length) / (
                        pairs - 1 if pairs > 1 else 1
                    )
                    positions.append((x, -half_width, 0))  # Left
                    positions.append((x, half_width, 0))  # Right
            else:
                # Odd number of wheels - include a center wheel
                positions.append((0, 0, 0))  # Center wheel
                pairs = (num_wheels - 1) // 2
                for i in range(pairs):
                    x_offset = (i + 1) * half_length / pairs
                    positions.append((x_offset, -half_width, 0))  # Front left
                    positions.append((x_offset, half_width, 0))  # Front right
                    positions.append((-x_offset, -half_width, 0))  # Rear left
                    positions.append((-x_offset, half_width, 0))  # Rear right

            return positions[
                :num_wheels
            ]  # Ensure we return exactly num_wheels positions

    def _map_steering_to_drive_wheels(self, drive_positions):
        steering_wheels = []

        # Use heuristics based on drive type
        if self.drive_type == DriveType.ACKERMANN:
            # For Ackermann, assume front wheels steer
            # Find the wheels with largest x coordinate (assuming x is forward)
            x_coords = [pos[0] for pos in drive_positions]
            front_indices = sorted(range(len(x_coords)), key=lambda i: -x_coords[i])
            steering_wheels = front_indices[: len(self.cfg.steering_joint_names)]
        elif self.drive_type == DriveType.ROVER:
            # For rover, assume corner wheels steer
            if len(drive_positions) >= 6:
                # Typical 6-wheel rover has corners at 0, 1, 4, 5
                steering_wheels = [0, 1, 4, 5]
            else:
                # For other configurations, find the corner wheels
                x_coords = [pos[0] for pos in drive_positions]
                y_coords = [pos[1] for pos in drive_positions]

                # Find min/max x and y to identify corners
                min_x, max_x = min(x_coords), max(x_coords)
                min_y, max_y = min(y_coords), max(y_coords)

                for i, (x, y, _) in enumerate(drive_positions):
                    # If close to extreme x and y values, consider it a corner
                    if (abs(x - min_x) < 0.1 or abs(x - max_x) < 0.1) and (
                        abs(y - min_y) < 0.1 or abs(y - max_y) < 0.1
                    ):
                        steering_wheels.append(i)

        # Ensure we don't have more steering wheels than steering joints
        steering_wheels = steering_wheels[: len(self.cfg.steering_joint_names)]

        return steering_wheels


class DriveType(Enum):
    SKID_STEER = auto()
    ACKERMANN = auto()
    ROVER = auto()


class WheelConfig(BaseModel):
    position: Tuple[float, float, float]
    radius: float
    drive_joint_name: str
    steering_joint_name: Optional[str] = None
    steering_limit: Optional[float] = None


@configclass
class WheeledDriveActionCfg(ActionTermCfg):
    class_type: Type = WheeledDriveAction

    # Scaling factors
    scale_linear: float = 1.0
    scale_angular: float = 1.0

    # Drive configuration (auto-detected if None)
    drive_type: Optional[DriveType] = None

    # Joint configuration
    drive_joint_names: List[str] = []
    steering_joint_names: List[str] = []

    # Vehicle parameters
    wheelbase: Tuple[float, float] = (0.8, 0.6)  # (length, width)
    wheel_radius: float = 0.1
    wheelbase_mid: Optional[float] = None  # Middle wheels width (for rovers)

    # Velocity limits
    max_linear_velocity: Optional[float] = None
    max_angular_velocity: Optional[float] = None



from typing import Any, Dict, List, Mapping, Set, Tuple

from isaaclab.utils.dict import *  # noqa: F403
from isaaclab.utils.dict import string_to_slice


def replace_slices_with_strings(data: Dict[Any, Any]) -> Dict[Any, Any]:
    """Replace slice objects with their string representations in a dictionary.

    Args:
        data: The dictionary to process.

    Returns:
        The dictionary with slice objects replaced by their string representations.
    """
    if isinstance(data, Mapping):
        return {k: replace_slices_with_strings(v) for k, v in data.items()}
    elif isinstance(data, Tuple):
        return tuple(replace_slices_with_strings(v) for v in data)
    elif isinstance(data, List):
        return [replace_slices_with_strings(v) for v in data]
    elif isinstance(data, Set):
        return {replace_slices_with_strings(v) for v in data}
    elif isinstance(data, slice):
        return f"slice({data.start},{data.stop},{data.step})"
    else:
        return data


def replace_strings_with_slices(data: Dict[Any, Any]) -> Dict[Any, Any]:
    """Replace string representations of slices with slice objects in a dictionary.

    Args:
        data: The dictionary to process.

    Returns:
        The dictionary with string representations of slices replaced by slice objects.
    """
    if isinstance(data, Mapping):
        return {k: replace_strings_with_slices(v) for k, v in data.items()}
    elif isinstance(data, Tuple):
        return tuple(replace_strings_with_slices(v) for v in data)
    elif isinstance(data, List):
        return [replace_strings_with_slices(v) for v in data]
    elif isinstance(data, Set):
        return {replace_strings_with_slices(v) for v in data}
    elif isinstance(data, str) and data.startswith("slice("):
        return string_to_slice(data)
    else:
        return data



from typing import Dict

import torch

from srb.core.env import ManipulationEnvVisualExtCfg, VisualExt
from srb.utils.cfg import configclass

from .task import Task, TaskCfg


@configclass
class VisualTaskCfg(ManipulationEnvVisualExtCfg, TaskCfg):
    def __post_init__(self):
        TaskCfg.__post_init__(self)
        ManipulationEnvVisualExtCfg.wrap(self, env_cfg=self)


class VisualTask(VisualExt, Task):
    cfg: VisualTaskCfg

    def __init__(self, cfg: VisualTaskCfg, **kwargs):
        Task.__init__(self, cfg, **kwargs)
        VisualExt.__init__(self, cfg, **kwargs)

    def _get_observations(self) -> Dict[str, torch.Tensor]:
        return {
            **Task._get_observations(self),
            **VisualExt._get_observations(self),
        }



from dataclasses import MISSING

from srb import assets
from srb.core.action import (
    InverseKinematicsActionGroup,
    JointPositionRelativeActionGroup,
    OperationalSpaceControlActionGroup,
)
from srb.core.asset import Articulation, AssetVariant, OrbitalManipulator
from srb.core.env.manipulation.env import (
    ManipulationEnv,
    ManipulationEventCfg,
    ManipulationSceneCfg,
)
from srb.core.env.mobile.orbital.env import (
    OrbitalEnv,
    OrbitalEnvCfg,
    OrbitalEventCfg,
    OrbitalSceneCfg,
)
from srb.core.manager import SceneEntityCfg
from srb.core.sensor import ContactSensorCfg
from srb.utils.cfg import configclass
from srb.utils.math import combine_frame_transforms_tuple


@configclass
class OrbitalManipulationSceneCfg(OrbitalSceneCfg, ManipulationSceneCfg):
    pass


@configclass
class OrbitalManipulationEventCfg(OrbitalEventCfg, ManipulationEventCfg):
    pass


@configclass
class OrbitalManipulationEnvCfg(OrbitalEnvCfg):
    ## Assets
    robot: OrbitalManipulator | AssetVariant = assets.GenericOrbitalManipulator(
        mobile_base=assets.Cubesat(), manipulator=assets.Franka()
    )
    _robot: OrbitalManipulator = MISSING  # type: ignore

    ## Scene
    scene: OrbitalManipulationSceneCfg = OrbitalManipulationSceneCfg()

    ## Events
    events: OrbitalManipulationEventCfg = OrbitalManipulationEventCfg()

    ## Time
    env_rate: float = 1.0 / 150.0
    agent_rate: float = 1.0 / 50.0

    def __post_init__(self):
        ## Jacobian-based actions are currently not supported for free-floating manipulators
        if isinstance(self.robot, OrbitalManipulator) and isinstance(
            self.robot.manipulator.actions,
            (InverseKinematicsActionGroup, OperationalSpaceControlActionGroup),
        ):
            self.robot.manipulator.actions = JointPositionRelativeActionGroup()

        super().__post_init__()
        assert self.scene.manipulator is not None

        ## Adapted from ManipulationEnvCfg
        # Sensor: End-effector transform
        self.scene.tf_end_effector.prim_path = f"{self.scene.manipulator.prim_path}/{self._robot.manipulator.frame_base.prim_relpath}"
        self.scene.tf_end_effector.target_frames[
            0
        ].prim_path = f"{self.scene.manipulator.prim_path}/{self._robot.manipulator.frame_flange.prim_relpath}"
        if self._robot.manipulator.end_effector is not None:
            (
                self.scene.tf_end_effector.target_frames[0].offset.pos,
                self.scene.tf_end_effector.target_frames[0].offset.rot,
            ) = combine_frame_transforms_tuple(
                self._robot.manipulator.frame_flange.offset.pos,
                self._robot.manipulator.frame_flange.offset.rot,
                self._robot.manipulator.end_effector.frame_tool_centre_point.offset.pos,
                self._robot.manipulator.end_effector.frame_tool_centre_point.offset.rot,
            )
        else:
            (
                self.scene.tf_end_effector.target_frames[0].offset.pos,
                self.scene.tf_end_effector.target_frames[0].offset.rot,
            ) = (
                self._robot.manipulator.frame_flange.offset.pos,
                self._robot.manipulator.frame_flange.offset.rot,
            )

        # Sensor: Robot contacts
        self.scene.contacts_robot.prim_path = f"{self.scene.manipulator.prim_path}/.*"

        # Sensor: End-effector contacts
        self.scene.contacts_end_effector = (
            ContactSensorCfg(
                prim_path=f"{self._robot.manipulator.end_effector.asset_cfg.prim_path}/.*",
            )
            if self._robot.manipulator.end_effector is not None
            else None
        )

        # Event: Randomize robot joints
        self.events.randomize_robot_joints.params["asset_cfg"] = SceneEntityCfg(
            "manipulator"
        )


class OrbitalManipulationEnv(OrbitalEnv, ManipulationEnv):
    cfg: OrbitalManipulationEnvCfg

    def __init__(self, cfg: OrbitalManipulationEnvCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._manipulator: Articulation = self.scene["manipulator"]



from srb.core.asset import Frame, RigidObjectCfg, Tool, Transform
from srb.core.sim import (
    CollisionPropertiesCfg,
    MassPropertiesCfg,
    MeshCollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.path import SRB_ASSETS_DIR_SRB_OBJECT


class ScoopRectangular(Tool):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/scoop",
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("scoop_rectangular.usdz").as_posix()
            ),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(mesh_approximation="sdf"),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(mass=0.001),
        ),
    )

    ## Frames
    frame_mount: Frame = Frame(prim_relpath="scoop")
    frame_tool_centre_point: Frame = Frame(offset=Transform(pos=(0.0, 0.0, 0.1)))


class ScoopSpherical(Tool):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/scoop",
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("scoop_spherical.usdz").as_posix()
            ),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(mesh_approximation="sdf"),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(mass=0.001),
        ),
    )

    ## Frames
    frame_mount: Frame = Frame(prim_relpath="scoop")
    frame_tool_centre_point: Frame = Frame(offset=Transform(pos=(0.0, 0.0, 0.1)))


class ScoopTriangular(Tool):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/scoop",
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("scoop_triangular.usdz").as_posix()
            ),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(mesh_approximation="sdf"),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(mass=0.001),
        ),
    )

    ## Frames
    frame_mount: Frame = Frame(prim_relpath="scoop")
    frame_tool_centre_point: Frame = Frame(offset=Transform(pos=(0.0, 0.0, 0.1)))



from isaaclab.envs import DirectMARLEnv as __DirectMARLEnv

from .cfg import DirectMarlEnvCfg


class DirectMarlEnv(__DirectMARLEnv):
    cfg: DirectMarlEnvCfg

    def __init__(self, cfg: DirectMarlEnvCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        # Apply visuals
        self.cfg.visuals.func(self.cfg.visuals)



from typing import Dict, Sequence, Tuple

import torch

from srb.core.env import InteractiveScene
from srb.utils.str import sanitize_cam_name

from .cfg import VisualExtCfg


class VisualExt:
    ## Subclass requirements
    scene: InteractiveScene

    def __init__(self, cfg: VisualExtCfg, **kwargs):
        self.__cameras = [
            (
                self.scene.sensors[camera_key],
                f"image_{sanitize_cam_name(camera_key)}",
                cfg.cameras_cfg[camera_key].data_types,
                cfg.cameras_cfg[camera_key].spawn.clipping_range,  # type: ignore
            )
            for camera_key in cfg.cameras_cfg.keys()
        ]

    def _get_observations(self) -> Dict[str, torch.Tensor]:
        return {
            image_key: image
            for camera, image_basename, data_types, clipping_range in self.__cameras
            for image_key, image in construct_observation(
                image_basename=image_basename,
                data_types=data_types,
                clipping_range=clipping_range,  # type: ignore
                **camera.data.output,
            ).items()
        }


def construct_observation(
    *,
    image_basename: str,
    data_types: Sequence[str],
    clipping_range: Tuple[float, float],
    merge_channels: bool = True,
    as_u8: bool = True,
    **kwargs,
) -> Dict[str, torch.Tensor]:
    processors = _PROCESSORS_U8 if as_u8 else _PROCESSORS_F32
    images = {
        f"{image_basename}_{data_type}": processors[data_type](
            kwargs[data_type], clipping_range
        )
        if data_type == "depth"
        else processors[data_type](kwargs[data_type])
        for data_type in data_types
    }
    if merge_channels:
        return {
            image_basename: torch.cat(
                [images[image_key] for image_key in images.keys()],
                dim=-1,
            )
        }
    else:
        return images


@torch.jit.script
def process_rgb_u8(image: torch.Tensor) -> torch.Tensor:
    return image[..., :3]


@torch.jit.script
def process_rgb_f32(image: torch.Tensor) -> torch.Tensor:
    return process_rgb_u8(image).to(torch.float32) / 255.0


@torch.jit.script
def process_depth_f32(
    image: torch.Tensor,
    clipping_range: Tuple[float, float],
) -> torch.Tensor:
    return (
        image.nan_to_num(
            nan=clipping_range[1], posinf=clipping_range[1], neginf=clipping_range[1]
        ).clamp(clipping_range[0], clipping_range[1])
        - clipping_range[0]
    ) / (clipping_range[1] - clipping_range[0])


@torch.jit.script
def process_depth_u8(
    image: torch.Tensor,
    clipping_range: Tuple[float, float],
) -> torch.Tensor:
    return (255.0 * process_depth_f32(image, clipping_range)).to(torch.uint8)


_PROCESSORS_U8 = {
    "rgb": process_rgb_u8,
    "depth": process_depth_u8,
}

_PROCESSORS_F32 = {
    "rgb": process_rgb_f32,
    "depth": process_depth_f32,
}



try:
    import rclpy
except ImportError:
    from srb.utils.ros import enable_ros2_bridge

    enable_ros2_bridge()

import os
import threading
from collections.abc import Callable

import numpy
import rclpy
from geometry_msgs.msg import Twist
from rclpy.executors import MultiThreadedExecutor
from rclpy.node import Node
from std_msgs.msg import Bool, Float64

from srb.interfaces.teleop import DeviceBase


class ROSTeleopInterface(DeviceBase, Node):
    def __init__(
        self,
        node: Node | None = None,
        pos_sensitivity: float = 1.0,
        rot_sensitivity: float = 3.1415927,
    ):
        if not node:
            rclpy.init()
            self._node = Node("sim", namespace="srb", start_parameter_services=False)  # type: ignore
        else:
            self._node = node

        self.pos_sensitivity = pos_sensitivity
        self.rot_sensitivity = rot_sensitivity

        self.sub_cmd_bel = self._node.create_subscription(
            Twist, "/cmd_vel", self.cb_twist, 1
        )
        self.sub_gripper = self._node.create_subscription(
            Bool, "/gripper", self.cb_event, 1
        )
        self.sub_latency = self._node.create_subscription(
            Float64, "/gui/latency", self.cb_latency, 1
        )

        self.latency = 0.0
        self.command_queue = []
        self.last_command = None

        # Command buffers
        self._close_gripper = False
        self._delta_pos = numpy.zeros(3)  # (x, y, z)
        self._delta_rot = numpy.zeros(3)  # (roll, pitch, yaw)

        # Run a thread for listening to device
        if not node:
            self._executor = MultiThreadedExecutor(num_threads=2)
            self._executor.add_node(self._node)
            self._thread = threading.Thread(target=self._executor.spin)
            self._thread.daemon = True
            self._thread.start()

    def cb_twist(self, msg):
        self._delta_pos[0] = self.pos_sensitivity * msg.linear.x
        self._delta_pos[1] = self.pos_sensitivity * msg.linear.y
        self._delta_pos[2] = self.pos_sensitivity * msg.linear.z

        self._delta_rot[0] = self.rot_sensitivity * msg.angular.x
        self._delta_rot[1] = self.rot_sensitivity * msg.angular.y
        self._delta_rot[2] = self.rot_sensitivity * msg.angular.z

    def cb_event(self, msg):
        if msg.data:
            self._close_gripper = not self._close_gripper

    def cb_latency(self, msg):
        if msg.data != self.latency:
            self.latency = msg.data
            self.command_queue = []
            self.feedback_queue = []
            self.last_command = None
            self.last_feedback = None

    def __del__(self):
        if hasattr(self, "_thread"):
            self._thread.join()

    def __str__(self) -> str:
        msg = f"ROS Interface ({self.__class__.__name__})\n"
        msg += f"Listenining on ROS_DOMAIN_ID: {os.environ.get('ROS_DOMAIN_ID', 0)}\n"
        return msg

    def reset(self):
        self._close_gripper = False
        self._delta_pos = numpy.zeros(3)
        self._delta_rot = numpy.zeros(3)
        self.command_queue = []
        self.feedback_queue = []
        self.last_command = None
        self.last_feedback = None

    def add_callback(self, key: str, func: Callable):
        raise NotImplementedError

    def advance(self) -> tuple[numpy.ndarray, bool]:
        commands = (
            numpy.concatenate([self._delta_pos, self._delta_rot]),
            self._close_gripper,
        )

        if self.latency == 0.0:
            return commands
        else:
            system_time = self._node.get_clock().now()
            self.command_queue.append((system_time, commands))

            # Find the last viable command
            last_viable_command = None
            for i, (t, _) in enumerate(self.command_queue):
                if (system_time - t).nanoseconds / 1e9 > self.latency:
                    last_viable_command = i
                else:
                    break

            if last_viable_command is not None:
                _, self.last_command = self.command_queue[last_viable_command]
                self.command_queue = self.command_queue[last_viable_command + 1 :]
                return self.last_command
            else:
                if self.last_command is not None:
                    return self.last_command
                else:
                    return numpy.zeros(6), commands[1]



from dataclasses import MISSING
from typing import Sequence, Tuple

import torch

from srb import assets
from srb._typing import StepReturn
from srb.core.action import ThrustAction
from srb.core.asset import AssetVariant, OrbitalRobot, Scenery
from srb.core.domain import Domain
from srb.core.env import OrbitalEnv, OrbitalEnvCfg, OrbitalEventCfg, OrbitalSceneCfg
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.marker import VisualizationMarkers, VisualizationMarkersCfg
from srb.core.mdp import push_by_setting_velocity  # noqa: F401
from srb.core.mdp import reset_root_state_uniform
from srb.core.sensor import ContactSensor, ContactSensorCfg
from srb.core.sim import PreviewSurfaceCfg, SphereCfg
from srb.utils.cfg import configclass
from srb.utils.math import deg_to_rad, matrix_from_quat, rotmat_to_rot6d

##############
### Config ###
##############


@configclass
class SceneCfg(OrbitalSceneCfg):
    env_spacing: float = 256.0

    ## Sensors
    contacts_robot: ContactSensorCfg = ContactSensorCfg(
        prim_path=MISSING,  # type: ignore
        track_air_time=True,
    )


@configclass
class EventCfg(OrbitalEventCfg):
    randomize_robot_state: EventTermCfg = EventTermCfg(
        func=reset_root_state_uniform,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("robot"),
            "pose_range": {
                "x": (-50.0, 50.0),
                "y": (-50.0, 50.0),
                "z": (100.0, 200.0),
                "roll": (-deg_to_rad(30.0), deg_to_rad(30.0)),
                "pitch": (-deg_to_rad(30.0), deg_to_rad(30.0)),
                "yaw": (-torch.pi, torch.pi),
            },
            "velocity_range": {
                "x": (-10.0, 10.0),
                "y": (-10.0, 10.0),
                "z": (-20.0, -10.0),
                "roll": (-deg_to_rad(15.0), deg_to_rad(15.0)),
                "pitch": (-deg_to_rad(15.0), deg_to_rad(15.0)),
                "yaw": (-deg_to_rad(15.0), deg_to_rad(15.0)),
            },
        },
    )
    # push_robot: EventTermCfg = EventTermCfg(
    #     func=push_by_setting_velocity,
    #     mode="interval",
    #     interval_range_s=(2.5, 10.0),
    #     params={
    #         "asset_cfg": SceneEntityCfg("robot"),
    #         "velocity_range": {
    #             "x": (-1.0, 1.0),
    #             "y": (-1.0, 1.0),
    #             "z": (-1.0, 1.0),
    #             "roll": (-deg_to_rad(10.0), deg_to_rad(10.0)),
    #             "pitch": (-deg_to_rad(10.0), deg_to_rad(10.0)),
    #             "yaw": (-deg_to_rad(10.0), deg_to_rad(10.0)),
    #         },
    #     },
    # )


@configclass
class TaskCfg(OrbitalEnvCfg):
    ## Scenario
    domain: Domain = Domain.MOON

    ## Assets
    robot: OrbitalRobot | AssetVariant = assets.PeregrineLander()
    scenery: Scenery | AssetVariant | None = AssetVariant.PROCEDURAL

    ## Scene
    scene: SceneCfg = SceneCfg()
    stack: bool = True

    ## Events
    events: EventCfg = EventCfg()

    ## Time
    episode_length_s: float = 40.0
    is_finite_horizon: bool = True

    ## Target
    tf_pos_target: Tuple[float, float, float] = (0.0, 0.0, 0.0)
    tf_quat_target: Tuple[float, float, float, float] = (1.0, 0.0, 0.0, 0.0)
    target_marker_cfg: VisualizationMarkersCfg = VisualizationMarkersCfg(
        prim_path="/Visuals/target",
        markers={
            "target": SphereCfg(
                radius=0.5,
                visual_material=PreviewSurfaceCfg(emissive_color=(0.2, 0.2, 0.8)),
            )
        },
    )

    def __post_init__(self):
        super().__post_init__()

        # Sensor: Robot contacts
        self.scene.contacts_robot.prim_path = f"{self.scene.robot.prim_path}/.*"


############
### Task ###
############


class Task(OrbitalEnv):
    cfg: TaskCfg

    def __init__(self, cfg: TaskCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._contacts_robot: ContactSensor = self.scene["contacts_robot"]
        self._target_marker: VisualizationMarkers = VisualizationMarkers(
            self.cfg.target_marker_cfg
        )

        ## Initialize buffers
        self._tf_pos_target = self.scene.env_origins + torch.tensor(
            self.cfg.tf_pos_target, dtype=torch.float32, device=self.device
        ).repeat(self.num_envs, 1)
        self._tf_quat_target = torch.tensor(
            self.cfg.tf_quat_target, dtype=torch.float32, device=self.device
        ).repeat(self.num_envs, 1)

        ## Visualize target
        self._target_marker.visualize(self._tf_pos_target, self._tf_quat_target)

    def _reset_idx(self, env_ids: Sequence[int]):
        super()._reset_idx(env_ids)

    def extract_step_return(self) -> StepReturn:
        ## Get remaining fuel (if applicable)
        if self._thrust_action_term_key:
            thrust_action_term: ThrustAction = self.action_manager._terms[  # type: ignore
                self._thrust_action_term_key
            ]
            remaining_fuel = (
                thrust_action_term.remaining_fuel / thrust_action_term.cfg.fuel_capacity
            ).unsqueeze(-1)
        else:
            remaining_fuel = None

        return _compute_step_return(
            ## Time
            episode_length=self.episode_length_buf,
            max_episode_length=self.max_episode_length,
            truncate_episodes=self.cfg.truncate_episodes,
            ## Actions
            act_current=self.action_manager.action,
            act_previous=self.action_manager.prev_action,
            ## States
            # Root
            tf_pos_robot=self._robot.data.root_pos_w,
            tf_quat_robot=self._robot.data.root_quat_w,
            vel_lin_robot=self._robot.data.root_lin_vel_b,
            vel_ang_robot=self._robot.data.root_ang_vel_b,
            projected_gravity_robot=self._robot.data.projected_gravity_b,
            # Transforms (world frame)
            tf_pos_target=self._tf_pos_target,
            # Contacts
            contact_forces_robot=self._contacts_robot.data.net_forces_w,  # type: ignore
            contact_robot=self._contacts_robot.compute_first_contact(self.step_dt),
            # IMU
            imu_lin_acc=self._imu_robot.data.lin_acc_b,
            imu_ang_vel=self._imu_robot.data.ang_vel_b,
            # Fuel
            remaining_fuel=remaining_fuel,
        )


@torch.jit.script
def _compute_step_return(
    *,
    ## Time
    episode_length: torch.Tensor,
    max_episode_length: int,
    truncate_episodes: bool,
    ## Actions
    act_current: torch.Tensor,
    act_previous: torch.Tensor,
    ## States
    # Root
    tf_pos_robot: torch.Tensor,
    tf_quat_robot: torch.Tensor,
    vel_lin_robot: torch.Tensor,
    vel_ang_robot: torch.Tensor,
    projected_gravity_robot: torch.Tensor,
    # Transforms (world frame)
    tf_pos_target: torch.Tensor,
    # Contacts
    contact_forces_robot: torch.Tensor,
    contact_robot: torch.Tensor,
    # IMU
    imu_lin_acc: torch.Tensor,
    imu_ang_vel: torch.Tensor,
    # Fuel
    remaining_fuel: torch.Tensor | None,
) -> StepReturn:
    num_envs = episode_length.size(0)
    dtype = episode_length.dtype
    device = episode_length.device

    ############
    ## States ##
    ############
    ## Root
    tf_rotmat_robot = matrix_from_quat(tf_quat_robot)
    tf_rot6d_robot = rotmat_to_rot6d(tf_rotmat_robot)

    ## Transforms (world frame)
    # Robot -> Target
    tf_pos_robot_to_target = tf_pos_target - tf_pos_robot
    vertical_distance_to_target = torch.clamp_min(-tf_pos_robot_to_target[:, 2], 0.001)

    ## Contacts
    MAX_TOUCHDOWN_LINEAR_VELOCITY = 2.0
    MAX_TOUCHDOWN_ANGLE = 0.174533  # 10 degrees
    touchdown = contact_robot.any(dim=1)
    crash = touchdown & (
        (torch.norm(vel_lin_robot, dim=1) > MAX_TOUCHDOWN_LINEAR_VELOCITY)
        | (torch.acos(-projected_gravity_robot[:, 2]) > MAX_TOUCHDOWN_ANGLE)
    )
    landed = touchdown | (torch.norm(contact_forces_robot.mean(dim=1), dim=1) > 1.0)

    ## Fuel
    remaining_fuel = (
        remaining_fuel
        if remaining_fuel is not None
        else torch.ones((num_envs, 1), dtype=dtype, device=device)
    )

    #############
    ## Rewards ##
    #############
    # Penalty: Action rate
    WEIGHT_ACTION_RATE = -0.025
    penalty_action_rate = WEIGHT_ACTION_RATE * torch.sum(
        torch.square(act_current - act_previous), dim=1
    )

    # Penalty: Fuel consumption
    WEIGHT_FUEL_CONSUMPTION = -2.0
    penalty_fuel_consumption = WEIGHT_FUEL_CONSUMPTION * torch.square(
        1.0 - remaining_fuel.squeeze(-1)
    )

    # Penalty: Angular velocity
    WEIGHT_ANGULAR_VELOCITY = -0.25
    MAX_ANGULAR_VELOCITY_PENALTY = -5.0
    penalty_angular_velocity = torch.clamp_min(
        WEIGHT_ANGULAR_VELOCITY * torch.sum(torch.square(vel_ang_robot), dim=1),
        min=MAX_ANGULAR_VELOCITY_PENALTY,
    )

    # Penalty: Alignment with gravity
    WEIGHT_GRAVITY_ROTATION_ALIGNMENT = -1.0
    penalty_gravity_rotation_alignment = WEIGHT_GRAVITY_ROTATION_ALIGNMENT * (
        torch.sum(torch.square(projected_gravity_robot[:, :2]), dim=1)
        + torch.square(projected_gravity_robot[:, 2] + 1.0)
    )

    # Penalty: Horizontal linear velocity
    WEIGHT_HORIZONTAL_LINEAR_VELOCITY = -0.1
    penalty_horizontal_linear_velocity = WEIGHT_HORIZONTAL_LINEAR_VELOCITY * torch.sum(
        torch.abs(vel_lin_robot[:, :2]), dim=1
    )

    # Reward: Vertical linear velocity
    WEIGHT_VERTICAL_LINEAR_VELOCITY = 8.0
    TANH_STD_VERTICAL_LINEAR_VELOCITY = 1.0
    reward_vertical_linear_velocity = WEIGHT_VERTICAL_LINEAR_VELOCITY * (
        1.0
        - torch.tanh(
            torch.clamp_min(
                torch.square(vel_lin_robot[:, 2]) - vertical_distance_to_target, 0.0
            )
            / TANH_STD_VERTICAL_LINEAR_VELOCITY
        )
    )

    # Reward: Horizontal distance to target
    WEIGHT_HORIZONTAL_DISTANCE = 4.0
    TANH_STD_HORIZONTAL_DISTANCE = 10.0
    reward_horizontal_distance = WEIGHT_HORIZONTAL_DISTANCE * (
        1.0
        - torch.tanh(
            torch.norm(tf_pos_robot_to_target[:, :2], dim=-1)
            / TANH_STD_HORIZONTAL_DISTANCE
        )
    )

    # Reward: Distance to target (precision landing)
    WEIGHT_LANDING_DISTANCE = 12.0
    TANH_STD_LANDING_DISTANCE = 0.5
    reward_landing_distance = (
        touchdown.float()
        * WEIGHT_LANDING_DISTANCE
        * (
            1.0
            - torch.tanh(
                torch.norm(tf_pos_robot_to_target, dim=-1) / TANH_STD_LANDING_DISTANCE
            )
        )
    )

    # Reward: Alignment with gravity (precision landing)
    WEIGHT_LANDING_ANGLE = 8.0
    reward_landing_angle = (
        landed.float()
        * WEIGHT_LANDING_ANGLE
        * (1.0 - torch.tanh(torch.acos(-projected_gravity_robot[:, 2])))
    )

    # Reward: Velocity (precision landing)
    WEIGHT_LANDING_VELOCITY = 16.0
    TANH_STD_LANDING_VELOCITY = 0.01
    reward_landing_velocity = (
        landed.float()
        * WEIGHT_LANDING_VELOCITY
        * (
            1.0
            - torch.tanh(
                (torch.norm(vel_lin_robot, dim=-1) + torch.norm(vel_ang_robot, dim=-1))
                / TANH_STD_LANDING_VELOCITY
            )
        )
    )

    # Reward: Remaining fuel (upon landing)
    WEIGHT_LANDING_FUEL = 4.0
    reward_landing_fuel = (
        landed.float() * WEIGHT_LANDING_FUEL * remaining_fuel.squeeze(-1)
    )

    # Penalty: Crash
    WEIGHT_CRASH = -64.0
    penalty_crash = WEIGHT_CRASH * crash.float()

    ##################
    ## Terminations ##
    ##################
    # Termination: Crash
    termination_crash = crash
    # Termination: Robot far below the target
    termination_below_target = tf_pos_robot_to_target[:, 2] > 10.0
    # Termination
    termination = termination_crash | termination_below_target
    # Truncation
    truncation = (
        episode_length >= max_episode_length
        if truncate_episodes
        else torch.zeros(num_envs, dtype=torch.bool, device=device)
    )

    return StepReturn(
        {
            "state": {
                "tf_rot6d_robot": tf_rot6d_robot,
                "vel_lin_robot": vel_lin_robot,
                "vel_ang_robot": vel_ang_robot,
                "projected_gravity_robot": projected_gravity_robot,
                "tf_pos_robot_to_target": tf_pos_robot_to_target,
            },
            "proprio": {
                "imu_lin_acc": imu_lin_acc,
                "imu_ang_vel": imu_ang_vel,
                "remaining_fuel": remaining_fuel,
            },
        },
        {
            "penalty_action_rate": penalty_action_rate,
            "penalty_fuel_consumption": penalty_fuel_consumption,
            "penalty_angular_velocity": penalty_angular_velocity,
            "penalty_gravity_rotation_alignment": penalty_gravity_rotation_alignment,
            "penalty_horizontal_linear_velocity": penalty_horizontal_linear_velocity,
            "reward_vertical_linear_velocity": reward_vertical_linear_velocity,
            "reward_horizontal_distance": reward_horizontal_distance,
            "reward_landing_distance": reward_landing_distance,
            "reward_landing_angle": reward_landing_angle,
            "reward_landing_velocity": reward_landing_velocity,
            "reward_landing_fuel": reward_landing_fuel,
            "penalty_crash": penalty_crash,
        },
        termination,
        truncation,
    )
